from si_fab import all as pdk
from ipkiss3 import all as i3
import pylab as plt
import numpy as np


def top_heated_waveguide(
    name,
    separation,
    offset_high,
    bend_radius,
    heated,
    heater_width,
    m1_length,
    trace_template,
    phase_error_width_deviation,
    phase_error_correlation_length,
):
    """Function that returns the top waveguide needed to complete the (heated) ring resonator.
    The waveguide can either be heated (si_fabs' HeatedWaveguide) or not.

    Parameters
    ----------
    name : str
    separation : float
    offset_high : float
    bend_radius : float
    heated : bool
    heater_width : float
    m1_length : float
    trace_template : TraceTemplate
    phase_error_width_deviation : float
    phase_error_correlation_length : float

    Returns
    -------
    cell : Waveguide or HeatedWaveguide
        U-shaped (heated) waveguide.

    """
    tt_new = trace_template.cell.modified_copy()
    tt_new.CircuitModel(
        phase_error_width_deviation=phase_error_width_deviation,
        phase_error_correlation_length=phase_error_correlation_length,
    )

    min_straight = bend_radius + m1_length
    shape = [
        (0, 0),
        (0.0, min_straight),
        (0.0, min_straight + offset_high),
        (separation, min_straight + offset_high),
        (separation, min_straight),
        (separation, 0.0),
    ]

    r_shape = i3.ShapeRound(original_shape=shape, radius=bend_radius)

    if heated:
        cell = pdk.HeatedWaveguide(
            name=name,
            trace_template=tt_new,
            m1_length=m1_length,
            heater_width=heater_width,
        )
    else:
        cell = i3.Waveguide(name=name, trace_template=tt_new)

    cell.Layout(shape=r_shape)

    return cell


class HeatedRing(i3.Circuit):
    """
    Heated ring resonator with the heater around the top waveguide.

    The parameters that describe this heated ring are:
    directional_coupler, m1_length, heater_width, center_wavelength and bend_radius.
    """

    directional_coupler = i3.ChildCellProperty(doc="Directional coupler")
    m1_length = i3.PositiveNumberProperty(default=10.0, doc="Length of the pad for the heater.")
    heater_width = i3.PositiveNumberProperty(default=0.6, doc="Width of the heater")
    center_wavelength = i3.PositiveNumberProperty(default=1.55, doc="Center wavelength")
    bend_radius = i3.PositiveNumberProperty(default=5.0, doc="Bend radius")
    phase_error_width_deviation = i3.NonNegativeNumberProperty(default=0.0, doc="Phase error width deviation")
    phase_error_correlation_length = i3.NonNegativeNumberProperty(default=0.0, doc="Phase error correlation length")

    def _default_directional_coupler(self):
        power_coupling = 0.5
        dir_coupler = pdk.SiDirectionalCouplerUPower(
            name=self.name + "dc",
            power_fraction=power_coupling,
            target_wavelength=self.center_wavelength,
        )
        return dir_coupler

    def _default_insts(self):
        instances = dict()
        instances["dc"] = self.directional_coupler
        tt = self.directional_coupler.get_default_view(i3.LayoutView).ports[0].trace_template
        distance = (
            self.directional_coupler.get_default_view(i3.LayoutView).ports["out2"].position.x
            - self.directional_coupler.get_default_view(i3.LayoutView).ports["in2"].position.x
        )

        wav = top_heated_waveguide(
            separation=distance,
            offset_high=2 * self.bend_radius,
            bend_radius=self.bend_radius,
            heated=True,
            m1_length=self.m1_length,
            trace_template=tt,
            phase_error_width_deviation=self.phase_error_width_deviation,
            phase_error_correlation_length=self.phase_error_correlation_length,
            heater_width=self.heater_width,
            name=self.name + "wg",
        )

        instances["wav"] = wav
        return instances

    def _default_specs(self):
        specs = [
            i3.Place("dc", (0, 0)),
            i3.Join("dc:in2", "wav:in"),
            i3.Join("dc:out2", "wav:out"),
        ]
        return specs

    def _default_exposed_ports(self):
        exposed_ports = {
            "dc:in1": "in",
            "dc:out1": "out",
            "wav:elec1": "ht_in",
            "wav:elec2": "ht_out",
        }
        return exposed_ports


def simulate_heated_ring(ring, heater_voltage, wavelengths, n_steps=10, debug=False):
    """Simulates the heated_ring and returns the S-matrix.

    Parameters
    ----------
    ring: HeatedRing
        HeatedRing cell
    heater_voltage: float
        heater voltage
    wavelengths : [floats]
        list of wavelengths at which we simulate
    n_steps : int
         number of time_steps before to stabilize the simulation
    debug : boolean

    Returns
    -------
    dict
        Dictionary with keys out1 and out2
    """

    # Define the function to create the sources' functions

    optical_in = i3.FunctionExcitation(port_domain=i3.OpticalDomain, excitation_function=lambda t: 1.0)
    gnd = i3.FunctionExcitation(port_domain=i3.ElectricalDomain, excitation_function=lambda t: 0.0)
    ht_v = i3.FunctionExcitation(port_domain=i3.ElectricalDomain, excitation_function=lambda t: heater_voltage)
    out_pb = i3.Probe(port_domain=i3.OpticalDomain)

    t0 = 0.0
    dt = 5e-9
    t1 = dt * n_steps

    testbench = i3.ConnectComponents(
        child_cells={
            "DUT": ring,
            "in": optical_in,
            "out": out_pb,
            "ht_in": ht_v,
            "ht_out": gnd,
        },
        links=[
            ("DUT:in", "in:out"),
            ("DUT:out", "out:in"),
            ("DUT:ht_in", "ht_in:out"),
            ("DUT:ht_out", "ht_out:out"),
        ],
    )

    cm = testbench.CircuitModel()
    s_out = []
    for w in wavelengths:
        res = cm.get_time_response(t0=t0, t1=t1, dt=dt, center_wavelength=w, debug=debug)
        s_out.append(res["out"][n_steps - 2])

    return {"out": s_out}


if __name__ == "__main__":
    cell = HeatedRing()
    lay = cell.Layout()
    lay.visualize(annotate=True)

    cm = cell.CircuitModel()
    wavelengths = np.linspace(1.54, 1.56, 200)
    S = cm.get_smatrix(wavelengths=wavelengths)
    S.visualize(term_pairs=[("in", "out")])

    heater_voltages = [0.0, 1.0]
    plt.figure()
    for hv in heater_voltages:
        print("simulating")
        results = simulate_heated_ring(ring=cell, heater_voltage=hv, wavelengths=wavelengths, debug=True)
        plt.plot(wavelengths, i3.signal_power(results["out"]), label=f"out1-{hv}")

    plt.legend()
    plt.show()
